#include "config.h"

#include <dlfcn.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "cache/cache.h"
#include "vsb.h"

#include "vmod_blob.h"
#include "vmod_blobdigest.h"
#include "tus_b64.h"
#include "tus_blob.h"

static const struct Vmod_vmod_blob_Func *vmod_blob = NULL;
static const struct Vmod_vmod_blobdigest_Func *vmod_blobdigest = NULL;
static void *dl_vmod_blob = NULL;
static void *dl_vmod_blobdigest = NULL;

/* ------------------------------------------------------------
 * linkage to blob and blobdigest vmods
 */

static void
load_vmod(const struct vmod_data *data, void **hdl, const void **func)
{
	const struct vmod_data *d;
	void *dlhdl;
	char buf[256];

	bprintf(buf, "./vmod_cache/_vmod_%s.%s",
	    data->name, data->file_id);

	dlhdl = dlopen(buf, RTLD_LAZY | RTLD_LOCAL);
	if (dlhdl == NULL) {
		fprintf(stderr, "dlopen vmod_%s: %s\n", data->name, dlerror());
		return;
	}
	bprintf(buf, "Vmod_%s_Data", data->name);
	d = dlsym(dlhdl, buf);
	if (d == NULL) {
		fprintf(stderr, "dlsym vmod_%s: %s\n", data->name, dlerror());
		(void) dlclose(dlhdl);
		return;
	}
	if (strcmp(d->name, data->name) ||
	    strcmp(d->file_id, data->file_id) ||
	    strcmp(d->abi, data->abi)) {
		fprintf(stderr, "vmod_%s: name/id/abi mismatch\n", data->name);
		(void) dlclose(dlhdl);
		return;
	}
	*hdl = dlhdl;
	*func = d->func;
	fprintf(stderr, "vmod_%s OK\n", data->name);
}

#define HENUM(t, b) static struct vmod_blobdigest_digest *digest_ ## t = NULL;
#include "tbl_hash_enum.h"

static unsigned enabled = 0;
static char *hashes = NULL;

int
tus_chksum_init(VRT_CTX) {
	struct vsb names[1];
	char buf[512];
	int first = 1;

	AN(VSB_new(names, buf, sizeof buf, VSB_FIXEDLEN));

	load_vmod(&import_Vmod_blob_Data, &dl_vmod_blob,
	    (const void **)&vmod_blob);
	load_vmod(&import_Vmod_blobdigest_Data, &dl_vmod_blobdigest,
	    (const void **)&vmod_blobdigest);

	if (vmod_blobdigest != NULL && vmod_blob != NULL)
		enabled = 1;
	else
		return (0);

	fprintf(stderr, "enabled %u\n", enabled);

#define HENUM(t, b) do {						\
		vmod_blobdigest->digest__init(ctx, &digest_ ## t,	\
		    "tus_digest_" #t, *vmod_blobdigest->enum_ ## b, NULL, \
		    *vmod_blobdigest->enum_TASK);			\
		if (digest_ ## t == NULL)				\
			break;						\
		if (first == 0)						\
			VSB_cat(names, ",");				\
		VSB_cat(names, #t);					\
		first = 0;						\
	} while (0);
#include "tbl_hash_enum.h"
	AZ(VSB_finish(names));
	REPLACE(hashes, VSB_data(names));
	fprintf(stderr, "digests OK\n");

	// dummy ref to appease compiler
	memset(&Vmod_vmod_blobdigest_Func, 0, sizeof Vmod_vmod_blobdigest_Func);
	memset(&Vmod_vmod_blob_Func, 0, sizeof Vmod_vmod_blob_Func);
	return (0);
}

int
tus_chksum_fini(VRT_CTX) {
#define HENUM(t, b) if (digest_ ## t != NULL)			\
		vmod_blobdigest->digest__fini(&digest_ ## t);
#include "tbl_hash_enum.h"
	REPLACE(hashes, NULL);
	if (dl_vmod_blob != NULL)
		(void) dlclose(dl_vmod_blob);
	if (dl_vmod_blobdigest != NULL)
		(void) dlclose(dl_vmod_blobdigest);
	return (0);
}

/* ------------------------------------------------------------
 * wrap base64
 */
VCL_BLOB
tus_b64_decode(VRT_CTX, const char *s, VCL_INT l)
{
	struct strands st = {.n = 1, .p = &s};

	if (! enabled)
		return (NULL);

	AN(vmod_blob);
	AN(vmod_blob->decode);
	AN(vmod_blob->enum_BASE64);

	return (vmod_blob->decode(ctx, *vmod_blob->enum_BASE64, l, &st));
}

/* ------------------------------------------------------------
 * enabled checksums
 */
const char *
tus_checksums(void)
{
	return (enabled ? hashes : NULL);
}

/* ------------------------------------------------------------
 * handling of tus checksums
 */

struct tus_chksum {
	unsigned			magic;
#define VMOD_TUS_CHKSUM_MAGIC		0x105c6650
	struct vmod_blobdigest_digest	*digest;
	VCL_BLOB			expect;
	VCL_BLOB			final;
};

/* tus hash names to blobdiget object */
struct vmod_blobdigest_digest *
tus_hash(const char *s, size_t l)
{
	if (enabled == 0 || s == NULL)
		return (NULL);

	if (l == 0)
		l = strlen(s);

#define HENUM(t, b)							\
	if (strncmp(s, #t, l) == 0) return (digest_ ## t);
#include "tbl_hash_enum.h"
	return (NULL);
}

struct tus_chksum *
tus_chksum_new(VRT_CTX, struct vmod_blobdigest_digest *d)
{
	struct tus_chksum *r;

	AN(d);

	r = WS_Alloc(ctx->ws, sizeof *r);
	if (r == NULL) {
		VRT_fail(ctx, "WS_Alloc failed");
		return (NULL);
	}
	INIT_OBJ(r, VMOD_TUS_CHKSUM_MAGIC);
	r->digest = d;

	return (r);
}

struct tus_chksum *
tus_chksum_hdr(VRT_CTX, const char *hdr)
{
	struct vmod_blobdigest_digest *d;
	struct tus_chksum *r;
	VCL_BLOB expect;
	const char *sep;

	if (enabled == 0 || hdr == NULL)
		return (NULL);

	sep = strchr(hdr, ' ');
	if (sep == NULL)
		return (NULL);

	d = tus_hash(hdr, (unsigned)(sep - hdr));
	if (d == NULL)
		return (NULL);

	r = tus_chksum_new(ctx, d);
	if (r == NULL)
		return (NULL);

	sep++;

	expect = tus_b64_decode(ctx, sep, 0);

	if (expect == NULL)
		return (NULL);

	r->expect = expect;
	return (r);
}

void
tus_chksum_update(VRT_CTX, struct tus_chksum *c,
    const void *ptr, size_t l)
{
	struct vrt_blob b;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(c, VMOD_TUS_CHKSUM_MAGIC);
	AN(enabled);

	b.type = 0x105;
	b.len = l;
	b.blob = ptr;

	(void) vmod_blobdigest->digest_update(ctx, c->digest, &b);
}

VCL_BLOB
tus_chksum_final(VRT_CTX, struct tus_chksum *c)
{

	AN(enabled);
	if (c->final == NULL)
		c->final = vmod_blobdigest->digest_final(ctx, c->digest);
	return (c->final);
}

VCL_BOOL
tus_chksum_equal(VRT_CTX, struct tus_chksum *c)
{
	(void) tus_chksum_final(ctx, c);
	return (vmod_blob->equal(ctx, c->expect, c->final));
}

// ------------------------------------------------------------
// not using vmod_blob because upload filenames should work without it

static inline char
hexnibble(unsigned char n)
{
	return (n < 0xa ? '0' + n : 'a' + n - 0xa);
}

VCL_STRING
tus_hex_buf(char * buf, size_t bufl, VCL_BLOB b)
{
	VCL_STRING r = buf;
	const unsigned char *p;
	unsigned char n;
	size_t l;

	if (b == NULL || b->len == 0)
		return ("");

	p = b->blob;
	l = b->len;
	while (l > 0) {
		if (bufl < 3)
			return (NULL);
		n = *p;
		buf[1] = hexnibble(n & 0xf);
		n >>= 4;
		buf[0] = hexnibble(n & 0xf);
		buf += 2;
		bufl -= 2;
		p++;
		l--;
	}
	AN(bufl);
	buf[0] = '\0';
	return (r);
}

VCL_STRING
tus_hex(VRT_CTX, VCL_BLOB b)
{
	char * buf;
	size_t bufl;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (b == NULL || b->len == 0)
		return ("");

	bufl = b->len * 2 + 1;
	buf = WS_Alloc(ctx->ws, bufl);
	if (buf == NULL)
		return (NULL);

	return (tus_hex_buf(buf, bufl, b));
}

void
tus_vsbhex(struct vsb *vsb, VCL_BLOB b)
{
	size_t l = b->len * 2 + 1;
	char buf[l];
	VSB_cat(vsb, tus_hex_buf(buf, l, b));
}
