#include "config.h"

#include <pthread.h>
#include <math.h>

#include "vdef.h"
#include "vrt.h"
#include "miniobj.h"
#include "vas.h"
#include "vbh.h"

#include "tus_file.h"
#include "tus_file_exp.h"

struct tus_exp {
	unsigned	magic;
#define TUS_EXP_MAGIC	0x105e8900
	unsigned	die;
	struct vbh	*heap;
	pthread_mutex_t	mtx;
	pthread_cond_t	cond;
	pthread_t	thread;
};

static inline vtim_real
fcore_when(const void *p)
{
	const struct tus_file_core *fcore;
	const struct tus_file_disk *fdisk;

	CAST_OBJ_NOTNULL(fcore, p, VMOD_TUS_FILE_CORE_MAGIC);
	fdisk = fcore->disk;
	CHECK_OBJ_NOTNULL(fdisk, VMOD_TUS_FILE_DISK_MAGIC);
	return (fdisk->upload_expires);
}

void *
tus_exp_thread(void *p)
{
	struct VPFX(tus_server) *srv;
	struct tus_file_core *fcore;
	struct timespec ts;
	struct tus_exp *e;
	vtim_real now, when, t;

	CAST_OBJ_NOTNULL(e, p, TUS_EXP_MAGIC);
	pthread_mutex_lock(&e->mtx);
	while (! e->die) {
		fcore = VBH_root(e->heap);
		if (fcore == NULL) {
			pthread_cond_wait(&e->cond, &e->mtx);
			continue;
		}

		when = fcore_when(fcore);
		now = VTIM_real();
		if (when > now) {
			assert(when > 1e9);
			ts.tv_nsec = (long)(modf(when, &t) * 1e9);
			ts.tv_sec = (long)t;
			assert(ts.tv_nsec >= 0 && ts.tv_nsec < 999999999);
			errno = pthread_cond_timedwait(&e->cond, &e->mtx, &ts);
			assert(errno == 0 || errno == ETIMEDOUT ||
			    errno == EINTR);
			continue;
		}

		pthread_mutex_unlock(&e->mtx);

		srv = fcore->server;
		tus_server_lock(srv);
		AZ(pthread_mutex_lock(&fcore->mtx));
		tus_file_del(&fcore);
		tus_server_unlock(srv);

		pthread_mutex_lock(&e->mtx);
	}
	pthread_mutex_unlock(&e->mtx);
}

void
tus_exp_delete(const struct tus_file_core *fcore)
{
	struct tus_exp *e = tus_server_exp(fcore->server);

	CHECK_OBJ_NOTNULL(e, TUS_EXP_MAGIC);

	pthread_mutex_lock(&e->mtx);
	assert(fcore->exp_idx != VBH_NOIDX);
	VBH_delete(e->heap, fcore->exp_idx);
	pthread_mutex_unlock(&e->mtx);
}

void
tus_exp_insert(struct tus_file_core *fcore)
{
	struct tus_exp *e = tus_server_exp(fcore->server);

	CHECK_OBJ_NOTNULL(e, TUS_EXP_MAGIC);

	pthread_mutex_lock(&e->mtx);
	VBH_insert(e->heap, fcore);
	if (VBH_root(e->heap) == fcore)
		pthread_cond_signal(&e->cond);
	pthread_mutex_unlock(&e->mtx);
}

void
tus_exp_touch(const struct tus_file_core *fcore)
{
	struct tus_exp *e = tus_server_exp(fcore->server);

	CHECK_OBJ_NOTNULL(e, TUS_EXP_MAGIC);

	pthread_mutex_lock(&e->mtx);
	VBH_reorder(e->heap, fcore->exp_idx);
	pthread_mutex_unlock(&e->mtx);
}

static int
tus_exp_cmp(void *priv, const void *a, const void *b) {

	(void) priv;

	return (fcore_when(a) < fcore_when(b));
}

static void
tus_exp_update(void *priv, void *p, unsigned u)
{
	struct tus_file_core *fcore;

	(void) priv;

	CAST_OBJ_NOTNULL(fcore, p, VMOD_TUS_FILE_CORE_MAGIC);

	fcore->exp_idx = u;
}

struct tus_exp *
tus_file_exp_new(void)
{
	struct tus_exp *e;
	pthread_attr_t attr;

	ALLOC_OBJ(e, TUS_EXP_MAGIC);
	AN(e);

	e->heap = VBH_new(NULL, tus_exp_cmp, tus_exp_update);
	AN(e->heap);

	AZ(pthread_mutex_init(&e->mtx, NULL));
	AZ(pthread_cond_init(&e->cond, NULL));

	AZ(pthread_attr_init(&attr));
	AZ(pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN));
	AZ(pthread_create(&e->thread, &attr, tus_exp_thread, e));
	AZ(pthread_attr_destroy(&attr));
	return (e);
}

void
tus_file_exp_destroy(struct tus_exp **ep)
{
	struct tus_exp *e;

	TAKE_OBJ_NOTNULL(e, ep, TUS_EXP_MAGIC);
	AN(e->heap);
	AZ(VBH_root(e->heap));

	e->die = 1;
	AZ(pthread_cond_signal(&e->cond));
	AZ(pthread_join(e->thread, NULL));

	AZ(pthread_cond_destroy(&e->cond));
	AZ(pthread_mutex_destroy(&e->mtx));

	free(e);
}
